package de.hawhamburg.reachability.properties;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.google.gson.Gson;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.properties.json.JsonRAPropertiesChange;
import de.hawhamburg.reachability.properties.json.JsonRAPropertiesRequest;
import de.hawhamburg.reachability.properties.json.JsonRAPropertiesResult;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.ExternFileUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.XMLUtil;

/**
 * Main properties service class which handles property requests and changes via ActiveMQ.<br><br>
 * 
 * This class represents a separately running service thread which handles the communication related to property 
 * requests and changes with the ActiveMQ at runtime. It will provide the at runtime defined properties to user 
 * interfaces, other systems and serves as interface for property changes. On such changes, this class will provide the 
 * update of the known system configuration of all related components and ensures that the configuration will be 
 * persisted to files for later restarts. 
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class PropertyService extends    Thread
                             implements StopableObject {
    
    /**
     * Basic (user-defined) properties.
     */
    private Properties props                = null;
    
    /**
     * Determines if the method {@link #shutDown()} has been called. 
     */
    private boolean shutDown                = false;
    
    /**
     * Listener for property requests.
     */
    private LPSubscriber requestProps       = null;
    
    /**
     * Publisher for property results.
     */
    private LPPublisher resultProps         = null;
    
    /**
     * Subscriber for other property results.
     */
    private LPSubscriber resultPropsReader  = null;
    
    /**
     * Listener for property change requests.
     */
    private LPSubscriber changeProps        = null;
    
    /**
     * Properties type.
     */
    private ConfigType type                 = null;
    
    /**
     * Path to the XML-file where the parameter description is located.
     */
    private String parameterDescriptionFile = null;
    
    /**
     * Parameter description.
     */
    private Document parameterDescription   = null;
    
    /**
     * The worker which called this property service.
     */
    private PropertyServiceNotifiable agent = null;
    
    /**
     * Create a new property service.
     * 
     * @param theWorker     The {@link PropertyServiceNotifiable} object which set up this property service reasoner 
     *                      and might be updated at runtime.
     * @param supportedType The type of properties for which this service is primary responsible as {@link ConfigType}.
     */
    public PropertyService(final PropertyServiceNotifiable theWorker,
                           final ConfigType supportedType) {
        this.agent                      = theWorker;
        this.type                       = supportedType;
        this.parameterDescriptionFile   = null;
    }
    
    /**
     * Create a new property service with a well known external properties description file.
     * 
     * @param theWorker     The {@link PropertyServiceNotifiable} object which set up this property service reasoner 
     *                      and might be updated at runtime.
     * @param supportedType The type of properties for which this service is primary responsible as {@link ConfigType}.
     * @param parmDescFile  The path to the well known external properties description file as {@link String}.
     */
    public PropertyService(final PropertyServiceNotifiable theWorker,
                           final ConfigType supportedType,
                           final String parmDescFile) {
        this.agent                      = theWorker;
        this.type                       = supportedType;
        this.parameterDescriptionFile   = parmDescFile;
    }

    /**
     * Loads the currently defined properties by delegation to the agent.
     */
    private void loadProperties() {
        this.props = agent.loadCurrentProperties(this.type);
    }
    
    /**
     * Alias for {@link #loadProperties()}.
     */
    public final void reloadProperties() {
        loadProperties();
    }
    
    /**
     * Loads the currently defined properties descriptions.
     */
    private void loadPropertiesDescription() {
        if (this.parameterDescriptionFile != null) {
            if (ExternFileUtil.exists(this.parameterDescriptionFile)) {
                this.parameterDescription = XMLUtil.loadDocument(this.parameterDescriptionFile);
            }
        }
    }

    /**
     * Run as long as {@link #shutDown()} wasn't called.
     */
    public final void run() {
        init();
        
        while (!this.shutDown) {
            handleRequested();
            handleChangeRequest();
            handleOtherConfigurations();
        }
        
        end();
    }
    
    /**
     * Initializes the ActiveMQ connections for request (<i>REACHABILITY_CONFIG_REQUEST_TOPIC</i>), response 
     * (<i>REACHABILITY_CONFIG_RESULT_TOPIC</i>) and change (<i>REACHABILITY_CONFIG_CHANGE_TOPIC</i>).
     */
    private void init() {
        OutputUtil.info("PropertyService started for " + agent.getClass().getSimpleName()
                        + " representing '" + type + "'");
        
        // Fetch topic names
        String requestTopicName = MainConfig.get("REACHABILITY_CONFIG_REQUEST_TOPIC");
        String resultTopicName  = MainConfig.get("REACHABILITY_CONFIG_RESULT_TOPIC");
        String changeTopicName  = MainConfig.get("REACHABILITY_CONFIG_CHANGE_TOPIC");
        
        // Set up subscribers/publishers
        this.requestProps       = ConnectionUtil.getSubscriber(requestTopicName, this);
        this.resultProps        = ConnectionUtil.getPublisher(resultTopicName, this);
        this.resultPropsReader  = ConnectionUtil.getSubscriber(resultTopicName, this);
        this.changeProps        = ConnectionUtil.getSubscriber(changeTopicName, this);
    }
    
    /**
     * Handles a request for properties sent over the ActiveMQ request topic <i>REACHABILITY_CONFIG_REQUEST_TOPIC</i> 
     * by calling {@link #pushProperties(JsonRAPropertiesRequest)}.
     */
    private void handleRequested() {
        String json = ConnectionUtil.getMessageOnTopic(requestProps);
        if (json != null) {
            Gson g = new Gson();
            JsonRAPropertiesRequest msgObj = g.fromJson(json, JsonRAPropertiesRequest.class);
            if (msgObj.Type.equals(this.type) && !this.agent.isSilent()) {
                pushProperties(msgObj);
            }
        }
    }
    
    /**
     * Sends a {@link JsonRAPropertiesResult} JSON object on the result topic (<i>REACHABILITY_CONFIG_RESULT_TOPIC</i>) 
     * containing the requested properties which are determined by using {@link #getPropertiesAsMap(String)} and their 
     * descriptions determined by {@link #getPropertiesDescriptionMap()}.
     * 
     * @param reqObj        A {@link JsonRAPropertiesRequest} JSON object representing the request.
     */
    public final void pushProperties(final JsonRAPropertiesRequest reqObj) {
        JsonRAPropertiesResult msg = new JsonRAPropertiesResult();

        msg.Id                      = getID();
        msg.Version                 = getVersion();
        msg.Type                    = this.type;
        msg.ServiceStateReference   = getServiceStateReference();
        if (reqObj != null) {
            msg.RequestId   = reqObj.Id;
            msg.Properties  = getPropertiesAsMap(reqObj.Property);
        } else {
            msg.Properties  = getPropertiesAsMap("*ALL");
        }
        msg.PropertiesDescription = getPropertiesDescriptionMap();
        
        ConnectionUtil.setMessageOnTopic(resultProps, msg);
    }
    
    /**
     * Loads the required properties for the requested type by using {@link #loadProperties()} and fetches the
     * requested properties defined by the <i>reqProps</i> parameter.
     *             
     * @param reqProps      The requested property as {@link String} ("*ALL" is defined as wildcard for all properties).
     * @return              The properties and their values as {@link Map} of {@link String} to {@link String}.
     */
    private Map<String, String> getPropertiesAsMap(final String reqProps) {
        loadProperties();
        
        Map<String, String> result = new HashMap<String, String>();
        for (String key : this.props.stringPropertyNames()) {
            if (key.equals(reqProps) || "*ALL".equals(reqProps)) {
                result.put(key, this.props.getProperty(key));
            }
        }
        return result;
    }
    
    /**
     * Loads the properties descriptions from the external file and parses them out of the XML document.
     * 
     * @return              The properties description as {@link Map} of {@link String} (property) to 
     *                      {@link PropertyDescription} (description object of the property).
     */
    private Map<String, PropertyDescription> getPropertiesDescriptionMap() {
        loadPropertiesDescription();
        
        Map<String, PropertyDescription> result = new HashMap<String, PropertyDescription>();
        
        // Groups
        String xGroups = "/configuration/description/group";
        int gNr = 0;
        for (Element group : XMLUtil.elementListByXPath(this.parameterDescription, xGroups)) {
            String groupName = group.getAttribute("display");
            
            // Properties
            String xProps = xGroups + "[" + (++gNr) + "]/property";
            int pNr = 0;
            for (Element prop : XMLUtil.elementListByXPath(this.parameterDescription, xProps)) {
                String key          = prop.getAttribute("key");
                String displayName  = prop.getAttribute("display");
                
                // Values
                String xValues              = xProps + "[" + (++pNr) + "]" + "/values";
                DisplayType displayType     = null;
                Map<String, String> vals    = new HashMap<String, String>();
                for (Element values : XMLUtil.elementListByXPath(this.parameterDescription, xValues)) {
                    if (displayType == null) {
                        String[] split  = values.getAttribute("type").split(":");
                        displayType     = DisplayType.valueOf(split[0]);
                        
                        switch(displayType) {
                            case ENUM:
                                // Value
                                String xValue = xValues + "[1]/value";
                                for (Element value : XMLUtil.elementListByXPath(this.parameterDescription, xValue)) {
                                    String tmpVal = value.getTextContent();
                                    if (value.hasAttribute("display")) {
                                        vals.put(tmpVal, value.getAttribute("display"));
                                    } else {
                                        vals.put(tmpVal, tmpVal);
                                    }
                                }
                                break;
                            case INT:
                                if (split.length > 1) {
                                    if (split.length > 2) {
                                        vals.put("min", split[1]);
                                        vals.put("max", split[2]);
                                    } else {
                                        vals.put("min", split[1]);
                                        vals.put("max", String.valueOf(Integer.MAX_VALUE));
                                    }
                                } else {
                                    vals.put("min", String.valueOf(Integer.MIN_VALUE));
                                    vals.put("max", String.valueOf(Integer.MAX_VALUE));
                                }
                                break;
                            case AGENT_DEFINED:
                                vals = agent.getValuesForProperty(key);
                                break;
                            default:
                                vals = new HashMap<String, String>();
                        }
                    }
                }
                
                // Build entry
                result.put(key, new PropertyDescription(displayName, displayType, groupName, vals));
            }
        }
        
        return result;
    }

    /**
     * Handles a properties change requested on the ActiveMQ change topic (<i>REACHABILITY_CONFIG_CHANGE_TOPIC</i>). 
     * It's changing its local copies of these properties and informs the creator of this service by using 
     * {@link PropertyServiceNotifiable#updateProperties(Properties, Properties, ConfigType)} on {@link #agent}.
     */
    private void handleChangeRequest() {
        String json = ConnectionUtil.getMessageOnTopic(changeProps);
        if (json != null) {
            // String to JSON-Object conversion
            Gson g = new Gson();
            JsonRAPropertiesChange msgObj = g.fromJson(json, JsonRAPropertiesChange.class);
            
            // Correct service state reference?
            if (msgObj.ServiceStateReference.equals(getServiceStateReference())) {
                // Fix local properties
                loadProperties();
                Properties localProps   = new Properties();
                Properties changedProps = new Properties();
                
                // Transfer current props
                for (Map.Entry<Object, Object> prop : this.props.entrySet()) {
                    localProps.put(String.valueOf(prop.getKey()), String.valueOf(prop.getValue()));
                }
    
                // Update internal properties
                for (Map.Entry<String, String> prop : msgObj.Properties.entrySet()) {
                    if (!props.containsKey(prop.getKey()) || !props.get(prop.getKey()).equals(prop.getValue())) {
                        changedProps.put(prop.getKey(), prop.getValue());
                    }
                    props.put(prop.getKey(), prop.getValue());
                    localProps.setProperty(prop.getKey(), prop.getValue());
                }
    
                // Send properties to agent
                agent.updateProperties(localProps, changedProps,  msgObj.Type);
            }
        }
    }
    
    /**
     * There may arrive configuration results via the ActiveMQ which dosn't equal {@link #type}. These updates may be
     * important for the {@link #agent} as well, so they need to be forwarded to it by this method.
     */
    private void handleOtherConfigurations() {
        String json = ConnectionUtil.getMessageOnTopic(this.resultPropsReader);
        if (json != null) {
            // String to JSON-Object conversion
            Gson g = new Gson();
            JsonRAPropertiesResult msgObj = g.fromJson(json, JsonRAPropertiesResult.class);
            
            // Transfer to properties object
            if (!msgObj.Id.equals(this.getID())) {
                Properties localProps = new Properties();
                for (Map.Entry<String, String> prop : msgObj.Properties.entrySet()) {
                    localProps.setProperty(prop.getKey(), prop.getValue());
                }
                
                // Send properties to agent
                agent.updateProperties(localProps, localProps, msgObj.Type);
            }
        }
    }
    
    /**
     * Gracefully end this service.
     */
    private void end() {
        if (requestProps != null) {
            requestProps.disconnect();
        }
        if (resultProps != null) {
            resultProps.disconnect();
        }
        if (resultPropsReader != null) {
            resultPropsReader.disconnect();
        }
        if (changeProps != null) {
            changeProps.disconnect();
        }
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     * 
     * @return              <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }
    
    /**
     * Returns a service ID.
     * 
     * @return              The service ID containing the {@link ProjectInfo#REACHABILITY_AGENT_WORKER_NAME}, the 
     *                      services name and a pseudo random number.
     */
    private String getID() {
        return ProjectInfo.REACHABILITY_AGENT_WORKER_NAME + "_PropertyService_" + this.type;
    }
    
    /**
     * Returns the service state reference which entitles a change requests.
     * 
     * @return              The service state reference which is required for change requests.
     */
    private String getServiceStateReference() {
        return this.type + "_" + this.agent.getClass().getSimpleName();
    }
    
    /**
     * Returns the agents version.
     * 
     * @return              The agents version as {@link ProjectInfo#REACHABILITY_AGENT_VERSION}.
     */
    private String getVersion() {
        return ProjectInfo.REACHABILITY_AGENT_VERSION;
    }
}
