package de.hawhamburg.reachability.sensor;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * Factory to create and start new sensor adapters.<br><br>
 * 
 * This factory provides static methods to get and start {@link SensorAdapter} objects. A sensor adapter needs to be 
 * located within the sub-package "adapter" and to implement the {@link Sensor} interface. 
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class SensorAdapterFactory {
    
    /**
     * Private default constructor to prevent instance creation of this utility class.
     */
    private SensorAdapterFactory() { }
    
    /**
     * Loads the adapter specified by its class name from the sub-package "adapter" using
     * {@link #getSensorAdapter(String, ReasonerDimension)} and connects it afterwards using 
     * {@link SensorAdapter#connect()}. It will then return the {@link SensorAdapter} object. If the 
     * {@link SensorAdapter} could not be found or instantiated, <code>null</code> will be returned instead.
     * 
     * @param className     The {@link SensorAdapter}s class name as string.
     * @param dimension     The {@link ReasonerDimension} for which the sensor will supply values.
     * @return              The {@link SensorAdapter} on success, otherwise <code>null</code>.
     */
    public static SensorAdapter getConnectedSensorAdapter(final String className,
                                                          final ReasonerDimension dimension) {
        // Try to get the sensor adapter
        SensorAdapter adapter = SensorAdapterFactory.getSensorAdapter(className, dimension);
        
        // Connect adapter if found
        if (adapter != null) {
            adapter.connect();
        }
        
        // Return the result
        return adapter;
    }
    
    /**
     * Tries to instantiate a {@link SensorAdapter} by a supplied class name from the sub-package "adapter". If the 
     * {@link SensorAdapter} could not be found or instantiated, <code>null</code> will be returned instead.
     * 
     * @param className     The {@link SensorAdapter}s class name as string.
     * @param dimension     The {@link ReasonerDimension} for which the sensor will supply values.
     * @return              The {@link SensorAdapter} on success, <code>null</code> otherwise.
     */
    @SuppressWarnings({ "rawtypes", "unchecked" })
    public static SensorAdapter getSensorAdapter(final String className,
                                                 final ReasonerDimension dimension) {
        SensorAdapter adapter = null;
        
        Class sensorClass;
        try {
            Constructor sensorConstructor;
            sensorClass = Class.forName(SensorAdapterFactory.class.getPackage().getName() + ".adapter." + className);
            
            if (SensorAdapter.class.isAssignableFrom(sensorClass)) {
                sensorConstructor   = sensorClass.getConstructor(ReasonerDimension.class);
                adapter             = (SensorAdapter) sensorConstructor.newInstance(new Object[]{dimension});
            } else {
                OutputUtil.error("[ClassCastException] Object of type " + className
                                 + " couldn't be casted to SensorAdapter");
            }
            
        } catch (ClassNotFoundException e) {
            OutputUtil.error("[ClassNotFoundException] "    + e.getMessage());
        } catch (InstantiationException e) {
            OutputUtil.error("[InstantiationException] "    + e.getMessage());
        } catch (IllegalAccessException e) {
            OutputUtil.error("[IllegalAccessException] "    + e.getMessage());
        } catch (NoSuchMethodException e) {
            OutputUtil.error("[NoSuchMethodException] "     + e.getMessage());
        } catch (SecurityException e) {
            OutputUtil.error("[SecurityException] "         + e.getMessage());
        } catch (IllegalArgumentException e) {
            OutputUtil.error("[IllegalArgumentException] "  + e.getMessage());
        } catch (InvocationTargetException e) {
            OutputUtil.error("[InvocationTargetException] " + e.getMessage());
        }
        return adapter;
    }
}
